"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const shuffle = require("lodash.shuffle");
const constants_1 = require("../constants");
const errors_1 = require("../errors");
const utils_1 = require("../utils");
const new_list_1 = require("./new_list");
const tried_list_1 = require("./tried_list");
class PeerBook {
    constructor({ newListConfig: newListConfig, triedListConfig: triedListConfig, secret, }) {
        this._newPeers = new new_list_1.NewList(newListConfig
            ? newListConfig
            : {
                secret,
                numOfBuckets: constants_1.DEFAULT_NEW_BUCKET_COUNT,
                bucketSize: constants_1.DEFAULT_NEW_BUCKET_SIZE,
                peerType: utils_1.PEER_TYPE.NEW_PEER,
            });
        this._triedPeers = new tried_list_1.TriedList(triedListConfig
            ? triedListConfig
            : {
                secret,
                numOfBuckets: constants_1.DEFAULT_TRIED_BUCKET_COUNT,
                bucketSize: constants_1.DEFAULT_TRIED_BUCKET_SIZE,
                peerType: utils_1.PEER_TYPE.TRIED_PEER,
            });
    }
    get newPeers() {
        return this._newPeers.peerList;
    }
    get triedPeers() {
        return this._triedPeers.peerList;
    }
    get allPeers() {
        return [...this.newPeers, ...this.triedPeers];
    }
    getRandomizedPeerList(minimumPeerDiscoveryThreshold, maxPeerDiscoveryResponseLength) {
        const allPeers = [...this.newPeers, ...this.triedPeers];
        const min = Math.ceil(Math.min(maxPeerDiscoveryResponseLength, allPeers.length * 0.25));
        const max = Math.floor(Math.min(maxPeerDiscoveryResponseLength, allPeers.length * 0.5));
        const random = Math.floor(Math.random() * (max - min + 1) + min);
        const randomPeerCount = Math.max(random, Math.min(minimumPeerDiscoveryThreshold, allPeers.length));
        return shuffle(allPeers).slice(0, randomPeerCount);
    }
    getPeer(peerInfo) {
        const triedPeer = this._triedPeers.getPeer(peerInfo.peerId);
        if (triedPeer) {
            return triedPeer;
        }
        return this._newPeers.getPeer(peerInfo.peerId);
    }
    hasPeer(peerInfo) {
        return (this._triedPeers.hasPeer(peerInfo.peerId) ||
            this._newPeers.hasPeer(peerInfo.peerId));
    }
    addPeer(peerInfo) {
        if (this._triedPeers.getPeer(peerInfo.peerId)) {
            throw new errors_1.ExistingPeerError(peerInfo);
        }
        this._newPeers.addPeer(peerInfo);
    }
    updatePeer(peerInfo) {
        if (this._triedPeers.getPeer(peerInfo.peerId)) {
            return this._triedPeers.updatePeer(peerInfo);
        }
        if (this._newPeers.getPeer(peerInfo.peerId)) {
            return this._newPeers.updatePeer(peerInfo);
        }
        return false;
    }
    removePeer(peerInfo) {
        this._newPeers.removePeer(peerInfo);
        this._triedPeers.removePeer(peerInfo);
    }
    upgradePeer(peerInfo) {
        if (this._triedPeers.hasPeer(peerInfo.peerId)) {
            return true;
        }
        if (this._newPeers.hasPeer(peerInfo.peerId)) {
            this.removePeer(peerInfo);
            this._triedPeers.addPeer(peerInfo);
            return true;
        }
        return false;
    }
    downgradePeer(peerInfo) {
        if (this._newPeers.hasPeer(peerInfo.peerId)) {
            return this._newPeers.failedConnectionAction(peerInfo);
        }
        if (this._triedPeers.hasPeer(peerInfo.peerId)) {
            const failed = this._triedPeers.failedConnectionAction(peerInfo);
            if (failed) {
                this.addPeer(peerInfo);
            }
        }
        return false;
    }
}
exports.PeerBook = PeerBook;
//# sourceMappingURL=peer_book.js.map