"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const validator_1 = require("validator");
const constants_1 = require("../constants");
const errors_1 = require("../errors");
const _1 = require(".");
const IPV4_NUMBER = 4;
const IPV6_NUMBER = 6;
const validateNetworkCompatibility = (peerInfo, nodeInfo) => {
    if (!peerInfo.sharedState) {
        return false;
    }
    if (!peerInfo.sharedState.networkId) {
        return false;
    }
    return peerInfo.sharedState.networkId === nodeInfo.networkId;
};
const validateProtocolVersionCompatibility = (peerInfo, nodeInfo) => {
    if (!peerInfo.sharedState) {
        return false;
    }
    if (typeof peerInfo.sharedState.protocolVersion !== 'string') {
        return false;
    }
    const peerHardForks = parseInt(peerInfo.sharedState.protocolVersion.split('.')[0], 10);
    const systemHardForks = parseInt(nodeInfo.protocolVersion.split('.')[0], 10);
    return systemHardForks === peerHardForks && peerHardForks >= 1;
};
exports.validatePeerCompatibility = (peerInfo, nodeInfo) => {
    if (!validateNetworkCompatibility(peerInfo, nodeInfo)) {
        return {
            success: false,
            error: constants_1.INCOMPATIBLE_NETWORK_REASON,
        };
    }
    if (!validateProtocolVersionCompatibility(peerInfo, nodeInfo)) {
        return {
            success: false,
            error: constants_1.INCOMPATIBLE_PROTOCOL_VERSION_REASON,
        };
    }
    return {
        success: true,
    };
};
exports.validatePeerAddress = (ipAddress, wsPort) => {
    if ((!validator_1.isIP(ipAddress, IPV4_NUMBER) && !validator_1.isIP(ipAddress, IPV6_NUMBER)) ||
        !validator_1.isPort(wsPort.toString())) {
        return false;
    }
    return true;
};
exports.validatePeerInfo = (peerInfo, maxByteSize) => {
    if (!peerInfo) {
        throw new errors_1.InvalidPeerInfoError(`Invalid peer object`);
    }
    if (!peerInfo.ipAddress ||
        !peerInfo.wsPort ||
        !exports.validatePeerAddress(peerInfo.ipAddress, peerInfo.wsPort)) {
        throw new errors_1.InvalidPeerInfoError(`Invalid peer ipAddress or port for peer with ip: ${peerInfo.ipAddress} and wsPort ${peerInfo.wsPort}`);
    }
    const byteSize = _1.getByteSize(peerInfo);
    if (byteSize > maxByteSize) {
        throw new errors_1.InvalidPeerInfoError(`PeerInfo is larger than the maximum allowed size ${maxByteSize} bytes`);
    }
    return peerInfo;
};
exports.validateNodeInfo = (nodeInfo, maxByteSize) => {
    const byteSize = _1.getByteSize(nodeInfo);
    if (byteSize > maxByteSize) {
        throw new errors_1.InvalidNodeInfoError(`Invalid NodeInfo was larger than the maximum allowed ${maxByteSize} bytes`);
    }
    return;
};
exports.validatePeerInfoList = (rawBasicPeerInfoList, maxPeerInfoListLength, maxPeerInfoByteSize) => {
    if (!rawBasicPeerInfoList) {
        throw new errors_1.InvalidPeerInfoListError(constants_1.INVALID_PEER_INFO_LIST_REASON);
    }
    const { peers } = rawBasicPeerInfoList;
    if (Array.isArray(peers)) {
        if (peers.length === 0) {
            return [];
        }
        if (peers.length > maxPeerInfoListLength) {
            throw new errors_1.InvalidPeerInfoListError(constants_1.PEER_INFO_LIST_TOO_LONG_REASON);
        }
        const sanitizedPeerList = peers.map(peerInfo => exports.validatePeerInfo(_1.sanitizeIncomingPeerInfo(peerInfo), maxPeerInfoByteSize));
        return sanitizedPeerList;
    }
    throw new errors_1.InvalidPeerInfoListError(constants_1.INVALID_PEER_INFO_LIST_REASON);
};
exports.validateRPCRequest = (request) => {
    if (!request) {
        throw new errors_1.InvalidRPCRequestError('Invalid request');
    }
    const rpcRequest = request;
    if (typeof rpcRequest.procedure !== 'string') {
        throw new errors_1.InvalidRPCRequestError('Request procedure name is not a string');
    }
    return rpcRequest;
};
exports.validateProtocolMessage = (message) => {
    if (!message) {
        throw new errors_1.InvalidProtocolMessageError('Invalid message');
    }
    const protocolMessage = message;
    if (typeof protocolMessage.event !== 'string') {
        throw new errors_1.InvalidProtocolMessageError('Protocol message is not a string');
    }
    return protocolMessage;
};
//# sourceMappingURL=validate.js.map